#include "mesh.h"
#include <iostream>
#include <fstream>
#include <string>

Pcre Mesh::regex_face_simple("^f\\s+(\\d*)\\s+(\\d*)\\s+(\\d*)");
Pcre Mesh::regex_face_full("^f\\s+(\\d*)/\\d*/(\\d*)\\s+(\\d*)/\\d*/(\\d*)\\s+(\\d*)/\\d*/(\\d*)");

//helper function for computing normal from three vertices
Vec3f *ComputeNormal(const Vec3f &p1, const Vec3f &p2, const Vec3f &p3)
{
	Vec3f v12 = p2;
	v12 -= p1;
	Vec3f v23 = p3;
	v23 -= p2;
	Vec3f *normal = new Vec3f();
	Vec3f::Cross3(*normal,v12,v23);
	normal->Normalize();
	return normal;
}

//constructor
Mesh::Mesh()
{
	bbox = NULL;
}

//constructor
Mesh::Mesh(char *file)
{
	bbox = NULL;
	parse(file);
}


//destructor
Mesh::~Mesh()
{
	//unallocate the materials
	map<string, Material *>::iterator t = materials.begin();
	for(; t != materials.end(); t++)
	{
		delete t->second;
	}
}

void Mesh::parse(char* filename)
{
	if(filename == NULL)
	{
		cerr << "ERROR: Must specify an input file" << endl;
		exit(1);
	}

	ifstream file (filename);
	
	string type;
	float x, y, z;
	
	//hacky hacky for things without materials
	Material *mat = new Material();
	materials["(null)"] = mat;
	
	
	while (file >> type)
	{
		if (type[0] == '#' || type[0] == 's')
		{
			char restOfLine [1024];
			file.getline(restOfLine, 1024);
		}
		else if (type == "v")
		{
			file >> x >> y >> z;
			verts.push_back(new Vec3f(x, y, z));
			if(!bbox)
				bbox = new BoundingBox(*verts.back(),*verts.back());
			else
				bbox->Extend(*verts.back());
		}
		else if (type == "vn")
		{
			file >> x >> y >> z;
			normals.push_back(new Vec3f(x, y, z));
		}
		else if (type == "vt")
		{
			cerr << "We care not for texture coordinates!" << endl;
		}
		else if (type == "mtllib")
		{
			file >> type;
			parseMaterials(type);
		}
		else if (type == "usemtl")
		{
			file >> type;
			mat = materials[type];
		}
		else if (type == "f")
		{
			//get entire line
			file.seekg(-1, ios_base::cur);
			string line;
			getline(file, line);
			
			//run regexs on line to determine which form we're dealing with
			if(regex_face_simple.search(line))
			{
				//simple form: get vertices then compute the appropriate normal
				int vI [3];
				vI[0] = atoi(regex_face_simple.get_match(0).c_str()) - 1;
				vI[1] = atoi(regex_face_simple.get_match(1).c_str()) - 1;
				vI[2] = atoi(regex_face_simple.get_match(2).c_str()) - 1;
				
				//we push this onto another vector to prevent it from screwing up the
				//index of normals
				c_normals.push_back(ComputeNormal(*verts[vI[0]], *verts[vI[1]], *verts[vI[2]]));
				Vec3f *n = c_normals.back();
				
				Triangle tri (verts[vI[0]], n, verts[vI[1]], n, verts[vI[2]], n, mat);
				triangles.push_back(tri);
			}
			else if(regex_face_full.search(line))
			{
				int vI [3]; int nI [3];
				vI[0] = atoi(regex_face_full.get_match(0).c_str()) - 1;
				vI[1] = atoi(regex_face_full.get_match(2).c_str()) - 1;
				vI[2] = atoi(regex_face_full.get_match(4).c_str()) - 1;
				
				nI[0] = atoi(regex_face_full.get_match(1).c_str()) - 1;
				nI[1] = atoi(regex_face_full.get_match(3).c_str()) - 1;
				nI[2] = atoi(regex_face_full.get_match(5).c_str()) - 1;
				
				Triangle tri (verts[vI[0]], normals[nI[0]], verts[vI[1]], normals[nI[1]], verts[vI[2]], normals[nI[2]], mat);
				triangles.push_back(tri);
			}
		}
		else
		{
			cerr << "unknown type: " << type << endl;
			char restOfLine [1024];
			file.getline(restOfLine, 1024);
		}
	}
	
	//output statistics about the mesh
	cout << "MESH STATISTICS" <<endl;
	cout << "===============" <<endl;
	cout << "Vertices: " << verts.size() << endl;
	cout << "Faces: " << triangles.size() << endl << endl;
}

void Mesh::parseMaterials(string f)
{
	Material *m = NULL;
	string token;
	float x,y,z;
	char *garbage;

	garbage = new char[1024];
	
	ifstream file(f.c_str());
	
	while(file >> token)
	{
		if(token == "newmtl")
		{
			string name;
			file >> name;
			m = new Material();
			m->setName(name);
			materials[name] = m;
		}
		else if(token == "Ns")
		{
			file >> x;
			m->setExponent(x);
		}
		else if(token == "Ka")
		{
			file >> x >> y >> z;
			m->setAmbient(Vec3f(x,y,z));
		}
		else if(token == "Kd")
		{
			file >> x >> y >> z;
			m->setDiffuse(Vec3f(x,y,z));
		}
		else if(token == "Ks")
		{
			file >> x >> y >> z;
			m->setSpecular(Vec3f(x,y,z));
		}
		//tokens we ignore
		else if(token == "#" || token == "Ni" || token == "d" || token == "illum")
		{
			file.getline(garbage,1024);
		}
		else
		{
			cerr << "unknown token" << endl;
		}
	}
	
	delete [] garbage;
}

//renders the mesh to the display
void Mesh::render()
{
	Triangle *t;
	Material *m = NULL;
	for(unsigned int i=0; i < triangles.size(); ++i)
	{
		t = &triangles[i];
		
		if(t->mtl != m)
		{
			m = t->mtl;
			m->gl();
		}
		glBegin(GL_TRIANGLES);
			glNormal3f(t->norms[0]->x(),t->norms[0]->y(),t->norms[0]->z());
			glVertex3f(t->verts[0]->x(),t->verts[0]->y(),t->verts[0]->z());
		
			glNormal3f(t->norms[1]->x(),t->norms[1]->y(),t->norms[1]->z());
			glVertex3f(t->verts[1]->x(),t->verts[1]->y(),t->verts[1]->z());
		
			glNormal3f(t->norms[2]->x(),t->norms[2]->y(),t->norms[2]->z());
			glVertex3f(t->verts[2]->x(),t->verts[2]->y(),t->verts[2]->z());
		glEnd();
	}
}

